@@ -19,6 +19,8 @@ from daggerml_cli.repo import (
     Ctx,
     Dag,
     Error,
+    Executable,
+    Fn,
     FnDag,
     Import,
     Index,
@@ -26,7 +28,6 @@ from daggerml_cli.repo import (
     Node,
     Ref,
     Repo,
-    Resource,
     unroll_datum,
 )
 from daggerml_cli.topology import node_info, topology
@@ -303,7 +304,7 @@ def delete_cache(config, cache_key: Ref):
 
 def list_cache(config):
     with Cache(config.CACHE_PATH) as cache:
-        return list(cache)
+        return cache.list()
 
 
 def info_cache(config, cache_key: Ref):
@@ -362,35 +363,56 @@ def describe_node(config: "Config", node: Ref):
 
 
 def backtrack_node(config: "Config", node: Ref, *_keys: Union[str, int]) -> Ref:
+    def fn(x, key):
+        if not isinstance(x, Ref) or x.type != "node" or not isinstance(x().data, Fn):
+            raise TypeError(f"invalid type {x!r} (type: {type(x().data)!s}) for backtrack_node")
+        argv = x().data.argv
+        fn, *args = (x().value().value for x in argv)
+        if fn.uri == "daggerml:list":
+            key = int(key)
+        elif fn.uri == "daggerml:dict":
+            i = args.index(key)
+            key = i + 1
+        elif fn.uri == "daggerml:get":
+            if key is not None:
+                raise Error(
+                    "cannot backtrack through 'get' with a non-null key",
+                    origin="dml",
+                    type="ValueError",
+                    stack=[
+                        {
+                            "filename": "api/invoke",
+                            "lineno": None,
+                            "function": "backtrack_node",
+                            "line": f"backtracking {fn.uri} with key: {key!r}",
+                        }
+                    ],
+                )
+            key = 0  # first arg is the collection
+        else:
+            raise Error(
+                "invalid collection type for backtrack_node",
+                origin="dml",
+                type="TypeError",
+                stack=[
+                    {
+                        "filename": "api/invoke",
+                        "lineno": None,
+                        "function": "backtrack_node",
+                        "line": fn.uri,
+                    }
+                ],
+            )
+        return argv[key + 1]
+
     with Repo.from_config(config) as db:
         with db.tx():
             keys = list(_keys)
+            if len(keys) == 0:
+                return fn(node, None)
             while len(keys) > 0:
                 key = keys.pop(0)
-                if not isinstance(node, Ref) or node.type != "node":
-                    raise TypeError(f"invalid type {node!r} for backtrack_node")
-                argv = node().data.argv
-                fn, *args = (x().value().value for x in argv)
-                if fn.uri == "daggerml:list":
-                    key = int(key)
-                elif fn.uri == "daggerml:dict":
-                    i = args.index(key)
-                    key = i + 1
-                else:
-                    raise Error(
-                        "invalid collection type for backtrack_node",
-                        origin="dml",
-                        type="TypeError",
-                        stack=[
-                            {
-                                "filename": "api/invoke",
-                                "lineno": None,
-                                "function": "backtrack_node",
-                                "line": fn.uri,
-                            }
-                        ],
-                    )
-                node = argv[key + 1]
+                node = fn(node, key)
             return node
 
 
@@ -430,27 +452,39 @@ def op_start_fn(db, index, argv, name=None, doc=None):
 @invoke_op
 def op_put_literal(db, index, data, name=None, doc=None):
     # TODO: refactor so that Resource.data -> Ref(datum)
-    def fn(args):
+    def maybe_to_node(args):
         fn_ = None
         if isinstance(args, list):
-            args = [fn(x) for x in args]
-            if any(isinstance(x, Ref) for x in args):
-                fn_ = db.put_datum(Resource("daggerml:list"))
+            args = [maybe_to_node(x) for x in args]
+            if any(isinstance(x, Ref) for x in args):  # only insert if needed
+                fn_ = db.put_datum(Executable("daggerml:list"))
         elif isinstance(args, dict):
-            args = {k: fn(v) for k, v in args.items()}
-            if any(isinstance(x, Ref) for x in args.values()):
-                fn_ = db.put_datum(Resource("daggerml:dict"))
+            args = {k: maybe_to_node(v) for k, v in args.items()}
+            if any(isinstance(x, Ref) for x in args.values()):  # only insert if needed
+                fn_ = db.put_datum(Executable("daggerml:dict"))
                 args = flatten(args.items())
         elif isinstance(args, set):
-            args = {fn(x) for x in args}
-            if any(isinstance(x, Ref) for x in args):
-                fn_ = db.put_datum(Resource("daggerml:set"))
+            args = {maybe_to_node(x) for x in args}
+            if any(isinstance(x, Ref) for x in args):  # only insert if needed
+                fn_ = db.put_datum(Executable("daggerml:set"))
+        elif isinstance(args, Executable):
+            attrs = {}
+            for x in ["data", "prepop"]:
+                attrs[x] = {}
+                for k, v in (getattr(args, x) or {}).items():
+                    attrs[x][k] = maybe_to_node(v)
+                    if isinstance(attrs[x][k], Ref) and attrs[x][k].type == "node":
+                        attrs[x][k] = db.get(attrs[x][k]).value
+                    else:
+                        attrs[x][k] = db.put_datum(attrs[x][k])
+            args = Executable(args.uri, adapter=args.adapter, **attrs)  # so we don't mutate
+            return args
         if fn_ is not None:
             return op_start_fn(db, index, [fn_, *args])
         return args
 
     with db.tx(True):
-        data = fn(data)
+        data = maybe_to_node(data)
         if isinstance(data, Ref) and data.type == "node":
             return op_set_node(db, index, name, data) if name else data
         result = db.put_node(Literal(db.put_datum(data)), index=index, name=name, doc=doc)
@@ -551,7 +585,7 @@ def invoke_api(config, token, data):
             op, args, kwargs = data
             if op in BUILTIN_FNS:
                 with db.tx(True):
-                    fn = db.put_datum(Resource(f"daggerml:{op}"))
+                    fn = db.put_datum(Executable(f"daggerml:{op}"))
                     fn = op_put_literal(db, index, fn, name=f"daggerml:{op}")
                 return op_start_fn(db, index, [fn, *args], **kwargs)
             return invoke_op.fns.get(op, no_such_op(op))(db, index, *args, **kwargs)
@@ -616,7 +650,6 @@ def commit_log_graph(config, output="ascii"):
                 return
             out = [*heads, *[x for x in reversed(nodes.values()) if x not in heads]]
             return out
-            # return print(json.dumps([asdict(x) for x in out], indent=2, sort_keys=True))
 
 
 def revert_commit(config, commit):
